#include "plugins-livido.h"
#include <stdio.h>
#include <string.h>

livido_setup_f livido_setup;

livido_init_f removal_init;
livido_deinit_f removal_deinit;
livido_process_f removal_process;

static uint8_t *av_luma_data;
static unsigned int av_count;

#define ABS(a)           (((a) < 0) ? -(a) : (a))

/* precomputed tables */
#define FP_BITS 18

static int Y_R[256];
static int Y_G[256];
static int Y_B[256];

static int Cb_R[256];
static int Cb_G[256];
static int Cb_B[256];
static int Cr_R[256];
static int Cr_G[256];
static int Cr_B[256];
static int conv_RY_inited = 0;


static int myround(double n)
{
  if (n >= 0) 
    return (int)(n + 0.5);
  else
    return (int)(n - 0.5);
}



static void init_RGB_to_YCbCr_tables(void)
{
  int i;

  /*
   * Q_Z[i] =   (coefficient * i
   *             * (Q-excursion) / (Z-excursion) * fixed-pogint-factor)
   *
   * to one of each, add the following:
   *             + (fixed-pogint-factor / 2)         --- for rounding later
   *             + (Q-offset * fixed-pogint-factor)  --- to add the offset
   *             
   */
  for (i = 0; i < 256; i++) {
    Y_R[i] = myround(0.299 * (double)i 
		     * 219.0 / 255.0 * (double)(1<<FP_BITS));
    Y_G[i] = myround(0.587 * (double)i 
		     * 219.0 / 255.0 * (double)(1<<FP_BITS));
    Y_B[i] = myround((0.114 * (double)i 
		      * 219.0 / 255.0 * (double)(1<<FP_BITS))
		     + (double)(1<<(FP_BITS-1))
		     + (16.0 * (double)(1<<FP_BITS)));

    Cb_R[i] = myround(-0.168736 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cb_G[i] = myround(-0.331264 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cb_B[i] = myround((0.500 * (double)i 
		       * 224.0 / 255.0 * (double)(1<<FP_BITS))
		      + (double)(1<<(FP_BITS-1))
		      + (128.0 * (double)(1<<FP_BITS)));

    Cr_R[i] = myround(0.500 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cr_G[i] = myround(-0.418688 * (double)i 
		      * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cr_B[i] = myround((-0.081312 * (double)i 
		       * 224.0 / 255.0 * (double)(1<<FP_BITS))
		      + (double)(1<<(FP_BITS-1))
		      + (128.0 * (double)(1<<FP_BITS)));
  }
  conv_RY_inited = 1;
}


static uint8_t 
calc_luma (uint8_t *pixel) {
  return (Y_R[pixel[0]] + Y_G[pixel[1]]+ Y_B[pixel[2]]) >> FP_BITS;
}

static uint32_t fastrand_val;

inline uint32_t fastrand(void)
{
#define rand_a 1073741789L
#define rand_c 32749L

  return (fastrand_val= rand_a * fastrand_val + rand_c);
}

void fastsrand(uint32_t seed)
{
	fastrand_val = seed;
}

static inline void 
make_black (uint8_t *pixel) {
  pixel[0]=pixel[1]=pixel[2]=(uint8_t)0;
}

/* declare our init function */
int common_init(livido_instance_t *inst) {
  av_luma_data=(uint8_t *)malloc(inst->out_channels[0].rowstrides[0]*inst->out_channels[0].height*sizeof(uint8_t));
  av_count=0;
  return 0;
}


static int common_process
(int type, livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  // host should have returned one input channel and one output channel
  // (which can be NULL, since we exported the property INPLACE)

  // we also set our input/output channels same as, so we can assume both frame
  // sizes and palettes match

  int i,j,k;
  int rowstride_i=env->in_channels[0].rowstrides[0];
  int rowstride_o;
  int width=env->in_channels[0].width*3;
  int size=env->in_channels[0].height*rowstride_i;

  uint8_t *old_pixel_data=in[0]->pixel_data, *new_pixel_data;

  int inplace=(out==NULL||out[0]==NULL);

  uint8_t luma;
  uint8_t av_luma;
  int param1;
  uint8_t luma_threshold;

  memcpy(&param1,env->parameter[0],sizeof(int));
  luma_threshold=(uint8_t)param1;

  if (inplace) {
    new_pixel_data=old_pixel_data;
    rowstride_o=rowstride_i;
  }
  else {
    new_pixel_data=out[0]->pixel_data;
    rowstride_o=env->out_channels[0].rowstrides[0];      
  }
  av_count++;
      
  for (i=0;i<size;i+=rowstride_i) {
    k=i/rowstride_i*rowstride_o;
    for (j=0;j<width;j+=3) {

      luma=calc_luma (&old_pixel_data[i+j]);
      if (av_count==1) {
	av_luma_data[k+j/3]=luma;
      }
      av_luma=(uint8_t)((double)(av_luma_data[k+j/3]*av_count)+(double)luma)/((double)av_count+1.);
      av_luma_data[k+j/3]=av_luma;
      if (ABS(luma-av_luma)>(luma_threshold)) {
	switch (type) {
	case 1:
	  // fire-ish effect
	  new_pixel_data[k+j]=((uint8_t)(((fastrand()&0xff)+(new_pixel_data[k+j+1]=((uint8_t) (fastrand()))))&0xff));     //R & G
	  new_pixel_data[k+j+2]=(uint8_t)0;                       //B
	  break;
	  //
	case 2:
	  break;
	case 0:
	  // make moving things black
	  make_black (&new_pixel_data[k+j]);
	}
      }
      else {
	// blue glow
	if (type==2) {
	  new_pixel_data[k+j]=new_pixel_data[k+j+1]=((uint8_t)(fastrand()&0xff));                                                 //R&G
	  new_pixel_data[k+j+2]=(uint8_t)255;  //B
	}}}}
  return 0;
}


int common_deinit (livido_instance_t *inst) {
  free (av_luma_data);
  return 0; // success
}

int t1_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  return common_process(0,env,in,out);
}

int t2_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  return common_process(1,env,in,out);
}

int t3_process
(livido_instance_t *env, livido_frame_t **in, livido_frame_t **out) {
  return common_process(2,env,in,out);
}

// create 1 input and one matching output channel, no masks

// allow RGB888 and BGR888 palettes

livido_palette_t in_palettes[]={LVD_PALETTE_BGR888,LVD_PALETTE_END};
livido_palette_t out_palettes[]={LVD_PALETTE_BGR888,LVD_PALETTE_END};

livido_channel_template_t in_channels[]={{"In 1",0,in_palettes,0,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

livido_channel_template_t out_channels[]={{"Out 1",LVD_CHNTMPL_SAME_AS_SIZE|LVD_CHNTMPL_SAME_AS_PALETTE,out_palettes,1,0,0},LIVIDO_CHANNEL_TEMPLATE_END};

static int paramdef1=200;
static int paramdef2=64;

livido_parameter_t params1[]={{"amount","blend amount",NULL,LVD_PARAM_NUMBER,0,0.,256.,&paramdef1,0,0},LIVIDO_PARAM_END};
livido_parameter_t params2[]={{"threshold","threshold amount",NULL,LVD_PARAM_NUMBER,0,0.,256.,&paramdef2,0,0},LIVIDO_PARAM_END};

livido_class_t fx[]={{"fg_bg_removal_type1","salsaman","fg/bg t1",1,1.7,LVD_PROPERTY_REALTIME|LVD_PROPERTY_INPLACE|LVD_PROPERTY_STATIC_CHANNEL_SIZES,params1,in_channels,out_channels,common_init,common_deinit,t1_process},{"fg_bg_removal_type2","salsaman","fg_bg_t2",1,1.7,LVD_PROPERTY_REALTIME|LVD_PROPERTY_INPLACE|LVD_PROPERTY_STATIC_CHANNEL_SIZES,params1,in_channels,out_channels,common_init,common_deinit,t2_process},{"fg_bg_removal_type3","salsaman","fg_bg_t3",1,1.7,LVD_PROPERTY_REALTIME|LVD_PROPERTY_INPLACE|LVD_PROPERTY_STATIC_CHANNEL_SIZES,params2,in_channels,out_channels,common_init,common_deinit,t3_process},LIVIDO_CLASS_END};

livido_class_t *livido_setup(void) {
  init_RGB_to_YCbCr_tables();
  return fx;
}

