// For official releases of livido.h, see www.dyne.org

// effects-livido.c
// LiVES
// (c) G. Finch 2004 <salsaman@xs4all.nl>
// Released under the GPL 2.0
// see file COPYING or www.fsf.org for licensing details 

// Unless otherwise specified these licensing details apply to all other files
// distributed as part of LiVES


#include "support.h"
#include "main.h"
#include "callbacks.h"

#ifdef USE_LIVIDO

#include "dlfcn.h"

// on a standard keyboard, ctrl-1 to ctrl-9 control effects, ctrl-0 clears fx
#define MAX_MODES_PER_KEY 8
#define MAX_LIVIDO_INSTANCES (FX_KEYS*MAX_MODES_PER_KEY)

// dummy channel size when no clip is open
#define DEFAULT_CHANNEL_HSIZE 320
#define DEFAULT_CHANNEL_VSIZE 240

// fixed size for generators - TODO - make configurable
#define DEF_GEN_WIDTH 640
#define DEF_GEN_HEIGHT 480

static livido_class_t *livido_fx[MAX_LIVIDO_PLUGINS];
static livido_instance_t *livido_instance[MAX_LIVIDO_INSTANCES];
static void *livido_handles[MAX_LIVIDO_PLUGINS];

// each 'hotkey' will control n instances, selectable as 'modes' or banks
static int key_to_instance[FX_KEYS>0?FX_KEYS-1:1][MAX_MODES_PER_KEY];
static int key_to_fx[FX_KEYS>0?FX_KEYS-1:1][MAX_MODES_PER_KEY];
static int key_modes[FX_KEYS>0?FX_KEYS-1:1];

static int next_free_instance=0;

gint num_livido_plugins=0;

// subtract this from num_livido_plugins to give physical number of plugins
static gint num_modes;

// arrays of in/out channels, these get created when the plugins are loaded
// each pointer is to a NULL terminated array of channels
static livido_channel_t *my_in_channels[MAX_LIVIDO_PLUGINS];
static livido_channel_t *my_out_channels[MAX_LIVIDO_PLUGINS];

// store keys so we now eg, which rte mask entry to xor when deiniting
static int fg_generator_key;
static int bg_generator_key;

// generators to start on playback, because of a problem in libvisual 
// we start generators after starting audio
static int bg_gen_to_start;
static int fg_gen_to_start;

// store the clip, this can sometimes get lost
static gint fg_generator_clip;

//#define DEBUG_LIVIDO

inline int get_next_free_instance(void) {
  int i;
  for (i=next_free_instance+1;i!=next_free_instance;i++) if (livido_instance[(i=(i==MAX_LIVIDO_INSTANCES?0:i))]==NULL) return i;
  return -1;
}

gboolean check_lvd_palette_list (livido_channel_template_t *chan_temp, livido_palette_t palette) {
  int i;

  if (chan_temp==NULL||chan_temp->name==NULL) return FALSE;
  if (palette==LVD_PALETTE_END) return TRUE;

  for (i=0;chan_temp->palettes[i]!=LVD_PALETTE_END;i++)
      if (chan_temp->palettes[i]==palette) return TRUE;
  return FALSE;
}


void set_in_channel_palettes (gint idx, gint num_channels) {
  // set in channel palettes for filter[idx]
  // in this case we fill first the mandatory channels,
  // then if necessary the optional channels
  int i;


  // fill mandatory channels
  for (i=0;my_in_channels[idx][i].template!=NULL;i++) {
    if (!my_in_channels[idx][i].template->flags&LVD_CHNTMPL_HINT_OPTIONAL) {
      my_in_channels[idx][i].flags=0; // indicate we will use it
      my_in_channels[idx][i].palette=LVD_PALETTE_BGR888;
      num_channels--;
    }
  }

  if (num_channels) {
    // OK we need to use some optional channels
    for (i=0;my_in_channels[idx][i].template!=NULL;i++) {
    if (my_in_channels[idx][i].template->flags&LVD_CHNTMPL_HINT_OPTIONAL) {
	// got one !
	if (num_channels--) {
	  // use this one
	  my_in_channels[idx][i].flags=0; // indicate we will use it
	  my_in_channels[idx][i].palette=LVD_PALETTE_BGR888;
	}
	else my_in_channels[idx][i].flags&=LVD_CHANNEL_DISABLED; // indicate we will not use it
      }
    }
  }
}




void set_out_channel_palettes (gint idx, gint num_channels) {
  // same as set in channels, but this time we cross ref the "same as"
  // palettes
  int i;

  // fill mandatory channels
  for (i=0;my_out_channels[idx][i].template!=NULL;i++) {
    if (!my_out_channels[idx][i].template->flags&LVD_CHNTMPL_HINT_OPTIONAL) {
      my_out_channels[idx][i].palette=LVD_PALETTE_BGR888;
      my_out_channels[idx][i].flags=0; // indicate we will use it
      num_channels--;
    }
  }

  if (num_channels) {
    // OK we need to use some optional channels
    for (i=0;my_out_channels[idx][i].template!=NULL;i++) {
      if (my_in_channels[idx][i].template->flags&LVD_CHNTMPL_HINT_OPTIONAL) {
	// got one !
	if (num_channels--) {
	  my_out_channels[idx][i].palette=LVD_PALETTE_BGR888;
	  my_out_channels[idx][i].flags=0;
	}
	else my_out_channels[idx][i].flags&=LVD_CHANNEL_DISABLED;
      }
    }
  }


  // now we set same as palettes
  for (i=0;my_in_channels[idx][i].template!=NULL;i++) {
    // in channels

    // palette
    if (my_in_channels[idx][i].template->same_as>0&&my_in_channels[idx][i].template->flags&LVD_CHNTMPL_SAME_AS_PALETTE) my_in_channels[idx][i].palette=my_in_channels[idx][my_in_channels[idx][i].template->same_as].palette;
  }

  for (i=0;my_out_channels[idx][i].template!=NULL;i++) {
    // in channels
    if (my_out_channels[idx][i].template->same_as>0&&my_out_channels[idx][i].template->flags&LVD_CHNTMPL_SAME_AS_PALETTE) my_out_channels[idx][i].palette=my_in_channels[idx][my_in_channels[idx][i].template->same_as].palette;
    else if (my_out_channels[idx][i].template->same_as<0&&my_out_channels[idx][i].template->flags&LVD_CHNTMPL_SAME_AS_PALETTE) my_out_channels[idx][i].palette=my_out_channels[idx][my_out_channels[idx][i].template->same_as].palette;

    // palette
  }

  // same_size will be checked in init function

}




gboolean check_for_lives(livido_class_t *filter, gint idx) {
  // for LiVES, currently:
  // all effects must take 0, 1 or 2 inputs and provide 
  // at least 1 output (for now)

  // all channels used must support BGR888 (for now)

  gint chans_in_mand=0;
  gint chans_in_opt_max=0;
  gint chans_out_mand=0;
  gint chans_out_opt_max=0;
  gboolean is_generator;
  
  int i;


  my_in_channels[idx]=my_out_channels[idx]=NULL;

  // for now we will only load realtime effects

  if (!(filter->properties&LVD_PROPERTY_REALTIME)) return FALSE;

  // count number of mandatory and optional in_channels
  // restrict it just to BGR888
  for (i=0;filter->in_channel_templates[i].name!=NULL;i++) {
    if (!filter->in_channel_templates[i].flags&LVD_CHNTMPL_HINT_OPTIONAL) {
      // is mandatory
      if (!check_lvd_palette_list (&filter->in_channel_templates[i],LVD_PALETTE_BGR888)) return FALSE;
      chans_in_mand++;
    }
    else {
      if (check_lvd_palette_list (&filter->in_channel_templates[i],LVD_PALETTE_BGR888)) chans_in_opt_max++;
    }
  }

  is_generator=!(chans_in_mand+chans_in_opt_max);

  // count number of mandatory and optional in_channels
  // restrict it just to BGR888
  for (i=0;filter->out_channel_templates[i].name!=NULL;i++) {
    if (!filter->out_channel_templates[i].flags&LVD_CHNTMPL_HINT_OPTIONAL) {
      // is mandatory
      if (!check_lvd_palette_list (&filter->out_channel_templates[i],LVD_PALETTE_BGR888)&&!(is_generator&&check_lvd_palette_list (&filter->out_channel_templates[i],LVD_PALETTE_RGB888))) return FALSE;
      chans_out_mand++;
    }
    else {
      if (check_lvd_palette_list (&filter->out_channel_templates[i],LVD_PALETTE_BGR888)||(is_generator&&check_lvd_palette_list (&filter->out_channel_templates[i],LVD_PALETTE_RGB888))) chans_out_opt_max++;
    }
  }

  if (chans_in_mand>2) return FALSE; // we dont handle mixers yet...
  if (chans_out_mand>1||!(chans_out_mand+chans_out_opt_max)) return FALSE;

  // create our copy of the channels
  my_in_channels[idx]=(livido_channel_t *)g_malloc ((chans_in_mand+chans_in_opt_max+1)*sizeof(livido_channel_t));
  my_out_channels[idx]=(livido_channel_t *)g_malloc ((chans_out_mand+chans_out_opt_max+1)*sizeof(livido_channel_t));
  
  for (i=0;i<chans_in_mand+chans_in_opt_max;i++) {
    my_in_channels[idx][i].template=&filter->in_channel_templates[i];
    my_in_channels[idx][i].shift_h=my_in_channels[idx][i].shift_v=0;
  }
  my_in_channels[idx][i].template=NULL;
  
  for (i=0;i<chans_out_mand+chans_out_opt_max;i++) {
    my_out_channels[idx][i].template=&filter->out_channel_templates[i];
    my_out_channels[idx][i].shift_h=my_out_channels[idx][i].shift_v=0;
  }
  my_out_channels[idx][i].template=NULL;

  return TRUE;
}


// TODO - don't use fix sized channels

void livido_init_all (void) {
// get list of plugins from directory and create our fx
  int i,j,idx=0,plugin_idx;
  gchar *plugin_name;
  GList *livido_fx_list=get_plugin_list(PLUGIN_LIVIDO_FX_BUILTIN,TRUE);
  gchar *plugin_path=g_strdup_printf("%s%s",PLUGIN_DIR,PLUGIN_LIVIDO_FX_BUILTIN);
  gchar *fx_path;
  gchar *msg;
  gint mode=0;
  gint skipped=0;
  gint key=-1;

  livido_setup_f *setup_fn;
  livido_class_t *fx_array;

#ifdef DEBUG_LIVIDO
  g_printerr("In livido init\n");
#endif
  num_modes=0;

  // danger Will Robinson !
  fg_gen_to_start=fg_generator_key=fg_generator_clip=-1;
  bg_gen_to_start=bg_generator_key=-1;

  // reset all instances to NULL
  for (i=0;i<MAX_LIVIDO_INSTANCES;livido_instance[i++]=NULL);
  for (i=0;i<FX_KEYS-1;i++) {
    key_modes[i]=0;
    for (j=0;j<MAX_MODES_PER_KEY;j++) {
      key_to_fx[i][j]=-1;
    }
  }

  // walk list and create fx structures
  for (plugin_idx=0;plugin_idx<g_list_length(livido_fx_list);plugin_idx++) {
    plugin_name=g_strdup(g_list_nth_data(livido_fx_list,plugin_idx));

    // TODO - replace with livido library fn
    fx_path=g_strdup_printf("%s/%s/%s",prefs->prefix_dir,plugin_path,plugin_name);

#ifdef DEBUG_LIVIDO
    g_printerr("Checking plugin %s\n",fx_path);
#endif

    if ((livido_handles[plugin_idx-skipped]=dlopen(fx_path,RTLD_LAZY))) {
      setup_fn=dlsym(livido_handles[plugin_idx-skipped],"livido_setup");
      mode=0;
      if (!(fx_array=(*setup_fn)())) g_printerr("bang, plugin exploded !\n");
      else {
	for (;fx_array->name!=NULL&&idx<MAX_LIVIDO_PLUGINS;fx_array++) {
	  if (check_for_lives(fx_array,idx)) {
	    if (!mode) key++;
	    if (key<FX_KEYS-1) msg=g_strdup_printf("Loaded effect \"%s\" in plugin \"%s\"; assigned to key ctrl-%d, mode %d.\n",fx_array->name,plugin_name,key+1,mode+1);
	    else msg=g_strdup_printf("Loaded effect \"%s\" in plugin \"%s\", no key assigned",fx_array->name,plugin_name);
	    d_print(msg);
	    g_free(msg);
	   
	    // TODO - load effects with variable number of in channels multiple
	    // times

	    if (my_in_channels[idx][0].template!=NULL) {
	      if (my_in_channels[idx][1].template==NULL) {
		// effect1
		set_in_channel_palettes (idx,1);
	      }
	      else {
		set_in_channel_palettes (idx,2);
	      }
	    }
	    set_out_channel_palettes (idx,1);
	    livido_fx[idx]=fx_array;
	    if (key<FX_KEYS-1) {
	      key_to_fx[key][mode]=idx;
	    }
	    mode++;
	    idx++;
	  }}}
      
      if (mode>0) num_modes+=--mode;
    }
    else {
      g_printerr(_ ("Unable to load plugin.\n"));
      skipped++;
    }
    g_free(plugin_name);
  }
  num_livido_plugins=idx-skipped;
  msg=g_strdup_printf(_ ("Loaded %d Livido effects\n"),idx-skipped);
  d_print(msg);
  g_free(msg);

  // TODO - add any rendered effects to fx submenu
}



GdkPixbuf *gdk_pixbuf_new_from_generator (void *src, GError **error) {
  GdkPixbuf *pixbuf=NULL;
  guchar *pixels;
  livido_frame_t *ofp[1];
  livido_instance_t *inst=(livido_instance_t *)src;
  livido_timecode_t tc=LIVIDO_TIMECODE_UNDEFINED;
  gint hsize,vsize,rowstride;

  if (inst==NULL) return NULL;
  
  // construct our gdk_pixbuf
  if ((pixbuf=gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, (hsize=inst->out_channels[0].width), (vsize=inst->out_channels[0].height)))==NULL) return NULL;

  pixels=gdk_pixbuf_get_pixels (pixbuf);

  // WARNING WARNING !
  if ((rowstride=gdk_pixbuf_get_rowstride (pixbuf)!=inst->out_channels[0].width*3)) return NULL;

  ofp[0]=(livido_frame_t *)g_malloc(sizeof(livido_frame_t));
  ofp[0]->timecode=tc;
  ofp[0]->x=0.;
  ofp[0]->y=0.;
  ofp[0]->pixel_data=pixels;
  ofp[0]->audio_data=NULL;
  ofp[0]->extra=NULL;
  (*inst->class->process)(inst,NULL,ofp);
  g_free (ofp[0]);
  
  if (inst->out_channels[0].palette==LVD_PALETTE_RGB888) {
    RGB2BGR_image (pixels,hsize,vsize,rowstride);
  }
  return pixbuf;
}

void livido_generator_end (livido_instance_t *inst) {
  // generator has stopped for one of the following reasons:
  // efect was de-inited; clip (bg/fg) was changed; playback stopped with fg
  gboolean is_bg=FALSE;
  gint current_file=mainw->current_file;

  if (inst==NULL) {
    g_printerr("  WARNING: inst was NULL !    ");
    return;
  }
  if (mainw->blend_file!=-1&&mainw->blend_file!=current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->ext_src==inst) is_bg=TRUE;
  
  if ((!is_bg&&mainw->rte_keys==fg_generator_key&&mainw->rte_keys!=bg_generator_key)||(is_bg&&mainw->rte_keys==bg_generator_key&&mainw->rte_keys!=fg_generator_key)) mainw->rte_keys=-1;
  if ((!is_bg&&mainw->last_grabable_effect==fg_generator_key&&mainw->last_grabable_effect!=bg_generator_key)||(is_bg&&mainw->last_grabable_effect==bg_generator_key&&mainw->last_grabable_effect!=fg_generator_key)) mainw->last_grabable_effect=-1;

  (*(inst->class)->deinit)(inst);

  LIVIDO_FREE_INSTANCE(inst->class,inst);
  inst=NULL;

  if (is_bg) {
    livido_instance[key_to_instance[bg_generator_key][key_modes[bg_generator_key]]]=NULL;
    key_to_instance[bg_generator_key][key_modes[bg_generator_key]]=-1;
    bg_gen_to_start=bg_generator_key=-1;
    mainw->current_file=mainw->blend_file;
    mainw->blend_file=-1;
  }
  else {
    livido_instance[key_to_instance[fg_generator_key][key_modes[fg_generator_key]]]=NULL;
    key_to_instance[fg_generator_key][key_modes[fg_generator_key]]=-1;
    if (mainw->rte&(1<<fg_generator_key)) mainw->rte^=(1<<fg_generator_key);
    fg_generator_key=fg_generator_clip=-1;
    if (mainw->blend_file==mainw->current_file) mainw->blend_file=-1;
  }
  cfile->ext_src=NULL;
  cfile->disk_frames--;

  cfile->end=cfile->frames=cfile->disk_frames;

  if (cfile->frames!=2) {
    if (cfile->disk_frames<1) {
      if (mainw->playing_file==-1) {
	close_current_file();
      }
      else {
	system (g_strdup_printf("smogrify close %s",cfile->handle));
	if (cfile->perf_counter!=0) {
	  g_free(cfile->events[EVENT_FRAMES]);
	  g_free(cfile->events[EVENT_FPS]);
	}
	g_free(cfile);
	cfile=NULL;
	
	if (mainw->first_free_file==-1||mainw->first_free_file>mainw->current_file) {
	  mainw->first_free_file=mainw->current_file;
	}
      }

      if (!is_bg) {
	if (mainw->blend_file==mainw->current_file) mainw->blend_file=-1;
	mainw->current_file=mainw->pre_src_file;
	if (mainw->playing_file>-1) {
	  mainw->reinit_effects=TRUE; // TODO
	  if (mainw->blend_file>-1&&mainw->blend_file!=current_file) {
	    mainw->pre_src_file=mainw->blend_file;

	    if (mainw->current_file==-1) mainw->current_file=mainw->pre_src_file;
	  }
	  if (mainw->pre_src_file!=-1&&mainw->files[mainw->pre_src_file]!=NULL)
	    do_quick_switch (mainw->pre_src_file);
	  else prevclip_callback(NULL,NULL,0,0,NULL);

	  if (mainw->whentostop==STOP_ON_VID_END) {
	    mainw->cancelled=1;
	    return;
	  }
	}
      }
      else {
	mainw->current_file=current_file;
	if (mainw->playing_file==-1) {
	  switch_to_file ((mainw->current_file=0),current_file);
	}
      }
      return;
    }
  
    // some frames were recorded
    g_snprintf(cfile->name,256,g_strdup_printf(_ ("Untitled%d"),mainw->untitled_number++));
    set_menu_text (cfile->menuentry,cfile->name,FALSE);
    add_to_winmenu();
    cfile->is_untitled=TRUE;
  }

  cfile->start=1;

  // TODO - note start and end times of each recorded section
  // record event_frame and event_fps then use quantise_events to bring it
  // all in time

  //quantise_events(1); // TODO - debug memory problems with this

  cfile->clip_type=CLIP_TYPE_DISK;
  cfile->changed=TRUE;
  cfile->bpp=24;
  if (is_bg) mainw->current_file=current_file;
  else if (mainw->playing_file==-1) {
    switch_to_file ((mainw->current_file=0),current_file);
  }
  return;
}

void livido_bg_generator_end (livido_instance_t *inst) {
  // when we stop with a bg generator we want it to be restarted next time
  int bg_gen_key=bg_generator_key;
  livido_generator_end (inst);
  bg_gen_to_start=bg_gen_key;
}


gboolean livido_generator_start (livido_instance_t *inst) {
  // make an "ephemeral clip"

  // cf. yuv4mpeg.c
  // start "playing" but receive frames from a plugin

  gint old_file=mainw->current_file;

  // create a virtual clip
  gint new_file=mainw->first_free_file;

  if ((mainw->current_file>-1&&cfile->clip_type==CLIP_TYPE_GENERATOR)||(mainw->blend_file!=-1&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR)) return FALSE;

  if (mainw->is_processing&&!mainw->preview) return FALSE;
  if ((mainw->preview||(mainw->current_file>-1&&cfile->opening))&&!mainw->num_tr_applied) return FALSE;


  if (old_file==-1||cfile->frames>0) {
    if (!get_new_handle(new_file,g_strdup_printf (_ ("generator:%s"),inst->class->name))) {
      mainw->error=TRUE;
      return FALSE;
    }

    if (mainw->playing_file==-1) mainw->pre_src_file=mainw->current_file;
    mainw->current_file=new_file;
    g_snprintf(cfile->type,40,"generator:%s",inst->class->name);
    cfile->achans=0;
    cfile->asampsize=0;
    
    // open as a clip with 1 frame
    cfile->start=cfile->end=cfile->frames=1;
    cfile->arps=cfile->arate=0;
    cfile->changed=FALSE;
  }
  else {
    new_file=old_file;
    cfile->frames=2;
  }


  cfile->clip_type=CLIP_TYPE_GENERATOR;
  cfile->ext_src=inst;
 
  if (!mainw->num_tr_applied) fg_generator_clip=new_file;

  cfile->disk_frames=0;
  cfile->pb_fps=cfile->fps=prefs->default_fps;

  cfile->hsize=inst->out_channels[0].width;
  cfile->vsize=inst->out_channels[0].height;

  cfile->opening=FALSE;
  cfile->proc_ptr=NULL;

  // allow clip switching
  cfile->is_loaded=TRUE;

  // record timecodes of frames - TODO - debug memory problems
  //cfile->perf_counter=cfile->perf_start=1;
  

  // if not playing, start playing
  if (mainw->playing_file==-1) {
    if (!mainw->num_tr_applied||old_file==-1) {
      switch_to_file((mainw->current_file=old_file),new_file);
      set_main_title(cfile->file_name,0);
      mainw->play_start=1;
      mainw->play_end=INT_MAX;
      if (mainw->num_tr_applied) {
	mainw->blend_file=mainw->current_file;
      }
    }
    else {
      mainw->blend_file=mainw->current_file;
      mainw->current_file=old_file;
      mainw->play_start=cfile->start;
      mainw->play_end=cfile->end;
      mainw->playing_sel=FALSE;
      if (mainw->rte&(1<<bg_generator_key)) mainw->rte^=(1<<bg_generator_key);
      
      if (mainw->last_grabable_effect==bg_generator_key+1) mainw->last_grabable_effect=EFFECT_NONE;
      if (mainw->rte_keys==bg_generator_key) mainw->rte_keys=-1;
    }
    if (mainw->play_window!=NULL&&old_file==-1) {
      // usually preview or load_preview_frame would do this
      g_signal_handler_block(mainw->play_window,mainw->pw_exp_func);
    }

    play_file();

    if (fg_generator_clip!=-1) {
      inst=mainw->files[fg_generator_clip]->ext_src;
      if (mainw->blend_file==-1||mainw->blend_file==mainw->current_file||mainw->blend_file!=new_file) {
	mainw->current_file=fg_generator_clip;
	livido_generator_end (inst);
      }
    }
    if (mainw->pre_src_file>-1) {
      switch_to_file ((mainw->current_file=0),mainw->pre_src_file);
      mainw->pre_src_file=-2;
    }
  }
  else {
    // already playing
    mainw->current_file=old_file;
    if (!mainw->num_tr_applied) {
      if (mainw->current_file==-1) mainw->current_file=new_file;
      do_quick_switch (new_file);
    }
    else {
      mainw->blend_file=new_file;
      if (mainw->rte&(1<<bg_generator_key)) mainw->rte^=(1<<bg_generator_key);
      mainw->whentostop=NEVER_STOP;
    }
  }
  return TRUE;
}






void set_channel_sizes (livido_channel_t *inc,livido_channel_t *outc) {
  if (inc[0].template!=NULL) {
    if (inc[1].template!=NULL) {
      if (mainw->current_file>-1) {
	// this is correct because we will always resize blend file to this
	inc[1].width=cfile->hsize;
	inc[1].height=cfile->vsize;
	inc[1].rowstrides[0]=3*cfile->hsize; // guess - TODO
      }
      else {
	inc[1].width=DEFAULT_CHANNEL_HSIZE;
	inc[1].height=DEFAULT_CHANNEL_VSIZE;
	inc[1].rowstrides[0]=DEFAULT_CHANNEL_HSIZE*3;
      }
    } 
    
    if (mainw->current_file>-1) {
      inc[0].width=cfile->hsize;
      inc[0].height=cfile->vsize;
      inc[0].rowstrides[0]=3*cfile->hsize;
    }
    else {
      mainw->reinit_effects=TRUE;
      inc[0].width=DEFAULT_CHANNEL_HSIZE;
      inc[0].height=DEFAULT_CHANNEL_VSIZE;
      inc[0].rowstrides[0]=DEFAULT_CHANNEL_HSIZE*3;
    }
      
    if (mainw->current_file>-1) {
      outc[0].width=cfile->hsize;
      outc[0].height=cfile->vsize;
      outc[0].rowstrides[0]=cfile->hsize*3;
    }
    else {
      outc[0].width=DEFAULT_CHANNEL_HSIZE;
      outc[0].height=DEFAULT_CHANNEL_VSIZE;
      outc[0].rowstrides[0]=DEFAULT_CHANNEL_HSIZE*3;
    }
  }
  else {
    outc[0].width=DEF_GEN_WIDTH;
    outc[0].height=DEF_GEN_HEIGHT;
    outc[0].rowstrides[0]=outc[0].width*3;
  }
}


gboolean livido_key_valid (int key) {
  if (key_to_fx[key][key_modes[key]]==-1) {
    return FALSE;
  }
  return TRUE;
}



gboolean livido_init_effect(int hotkey, gint rowstride) {
  livido_class_t *fx;
  livido_instance_t *instance;
  livido_channel_t *inc,*outc;
  gint num_tr_applied;
  gboolean fg_modeswitch=FALSE,is_trans=FALSE,is_gen=FALSE,trans_nextmode=FALSE;
  gint rte_keys=mainw->rte_keys;
  gint saved_rte=-1;

  if (hotkey<0) {
    trans_nextmode=TRUE;
    hotkey=-hotkey;
  }

  if (hotkey==fg_generator_key) {
    fg_modeswitch=TRUE;
  }

  if (!livido_key_valid (hotkey)) {
    return FALSE;
  }


  if (next_free_instance==-1) {
    d_print(_ ("Warning - no more effect slots left !\n"));
    return FALSE;
  }

  if (my_in_channels[key_to_fx[hotkey][key_modes[hotkey]]]->template==NULL) {
    // starting a new generator, we must stop the old one
    // warning = in future, bg *and* fg generators could change
    if (mainw->num_tr_applied&&mainw->blend_file!=-1&&mainw->blend_file!=mainw->current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR) {

      if (bg_gen_to_start==-1) {
	saved_rte=mainw->rte;
	livido_generator_end(mainw->files[mainw->blend_file]->ext_src);
      }
      bg_gen_to_start=bg_generator_key=-1;
      mainw->blend_file=-1;
    }
    if (mainw->current_file>0&&cfile->clip_type==CLIP_TYPE_GENERATOR) {
      if (mainw->noswitch||mainw->is_processing||mainw->preview) return FALSE; // stopping fg gen will cause clip to switch
      if (saved_rte==-1) saved_rte=mainw->rte;
      livido_generator_end (cfile->ext_src);
      mainw->reinit_effects=TRUE;
      fg_generator_key=fg_generator_clip=-1;
    }
  }

  fx=livido_fx[key_to_fx[hotkey][key_modes[hotkey]]];


  inc=my_in_channels[key_to_fx[hotkey][key_modes[hotkey]]];
  outc=my_out_channels[key_to_fx[hotkey][key_modes[hotkey]]];

  if (inc[0].template!=NULL) {
    if (inc[1].template!=NULL) {
      mainw->num_tr_applied++; // increase trans count
      if (mainw->num_tr_applied==1&&!trans_nextmode) {
	mainw->blend_file=mainw->current_file;
	mainw->blend_file_step=(mainw->current_file>-1&&cfile->pb_fps<0)?-1:1;
      }
    }
  }
  else {
    // aha - a generator
     if (mainw->playing_file==-1) {
      // if we are not playing, we will postpone creating the instance
      // this is a workaround for a problem in libvisual
      if (mainw->num_tr_applied) {
	bg_gen_to_start=hotkey;
      }
      else fg_gen_to_start=hotkey;
      is_gen=TRUE;
    }
     else if (!fg_modeswitch&&!mainw->num_tr_applied&&(mainw->noswitch||mainw->is_processing||mainw->preview)) return FALSE;
  }

  set_channel_sizes (inc,outc);
  LIVIDO_CREATE_INSTANCE_C(fx,instance,inc,outc);

  livido_instance[next_free_instance]=instance;
  key_to_instance[hotkey][key_modes[hotkey]]=next_free_instance;

  mainw->last_grabable_effect=hotkey;

  if (!is_gen) {
    if ((*fx->init)(instance)) {
      d_print (g_strdup_printf (_ ("Failed to start instance %s\n"),instance->class->name));
      LIVIDO_FREE_INSTANCE(fx,instance);
      livido_instance[key_to_instance[hotkey][key_modes[hotkey]]]=NULL;
      key_to_instance[hotkey][key_modes[hotkey]]=-1;
      if (is_trans) mainw->num_tr_applied--;
      return FALSE;
    }
    next_free_instance=get_next_free_instance();
  }

  if (inc[0].template==NULL) {
    // generator start
    if (mainw->num_tr_applied&&!fg_modeswitch) {
      // transition is on, make into bg clip
      bg_generator_key=hotkey;
    }
    else {
      // no transition, make fg (or kb was grabbed for fg generator)
      fg_generator_key=hotkey;
    }

    // start the generator and maybe start playing
    num_tr_applied=mainw->num_tr_applied;
    if (fg_modeswitch) mainw->num_tr_applied=0; // force to fg
    if (!livido_generator_start(instance)) {
      d_print (g_strdup_printf (_ ("Unable to start generator %s\n"),instance->class->name));
      if (mainw->num_tr_applied) {
	bg_gen_to_start=bg_generator_key=-1;
      }
      else {
	fg_generator_key=fg_generator_clip=-1;
      }
      mainw->num_tr_applied=num_tr_applied;
      return FALSE;
    }
    // TODO - problem if modeswitch triggers playback
    if (fg_modeswitch) mainw->num_tr_applied=num_tr_applied;
    if (fg_generator_key!=-1) mainw->rte|=(1<<fg_generator_key);
  }

  if (rte_keys==hotkey) mainw->rte_keys=rte_keys;

  return TRUE;
}



gboolean livido_playback_gen_start (void) {
  // init generators on pb. We have to do this after audio startup
  livido_instance_t *inst;

  if (bg_gen_to_start!=-1) {
    if (fg_gen_to_start==-1) {
      if (my_in_channels[key_to_fx[bg_gen_to_start][key_modes[bg_gen_to_start]]][0].template==NULL) { // check is still gen
	inst=livido_instance[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]];
	if (inst==NULL) {
	  gint playing_file=mainw->playing_file;
	  mainw->playing_file=-100; //kludge to stop playing a second time
	  if (livido_init_effect (bg_gen_to_start,cfile->hsize*3)==FALSE) {
	    mainw->playing_file=-1;
	    return FALSE;
	  }
	  mainw->playing_file=playing_file;
	  inst=livido_instance[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]];
	}
	else {
	  // this shouldn't happen
	  if ((*inst->class->init)(inst)) {
	    d_print (g_strdup_printf (_ ("warn: Failed to start generator %s\n"),inst->class->name));
	    (*inst->class->deinit)(inst);
	    LIVIDO_FREE_INSTANCE(inst->class,inst);
	    livido_instance[key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]]=NULL;
	    key_to_instance[bg_gen_to_start][key_modes[bg_gen_to_start]]=-1;
	    bg_gen_to_start=-1;
	    return FALSE;
	  }
	}
	mainw->files[mainw->blend_file]->ext_src=inst;
	next_free_instance=get_next_free_instance();
      }
    }
    bg_gen_to_start=-1;
  }

  if (fg_gen_to_start!=-1) {
    if (my_in_channels[key_to_fx[fg_gen_to_start][key_modes[fg_gen_to_start]]][0].template==NULL) { // check is still gen
      inst=livido_instance[key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]];
      if (inst!=NULL&&(*inst->class->init)(inst)) {
	d_print (g_strdup_printf (_ ("Warn: Failed to start generator %s\n"),inst->class->name));
	(*inst->class->deinit)(inst);
	LIVIDO_FREE_INSTANCE(inst->class,inst);
	livido_instance[key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]]=NULL;
	key_to_instance[fg_gen_to_start][key_modes[fg_gen_to_start]]=-1;
	fg_gen_to_start=-1;
	cfile->ext_src=NULL;
	return FALSE;
      }
      cfile->ext_src=inst;
      next_free_instance=get_next_free_instance();
      if (mainw->save_to_ext) toggle_record();
    }
    fg_gen_to_start=-1;
  }
  return TRUE;
}
 



#define KEYSCALE 256
 


gint livido_get_blend_factor(int hotkey) {

  livido_instance_t *instance;
  const livido_class_t *fx;
  int idx;

  if (hotkey<0) return 0;

  idx=key_to_instance[hotkey][key_modes[hotkey]];

  if (idx==-1||(instance=livido_instance[idx])==NULL) return 0;

  fx=instance->class;
  if (fx->parameters==NULL) return 0;
  switch (fx->parameters[0].type) {
  case LVD_PARAM_NUMBER:
  if (fx->parameters[0].min==fx->parameters[0].max) return (gint)fx->parameters[0].max;
    if (fx->parameters[0].decimals) {
      gfloat val=get_float_param(instance->parameter[0]);
      return (gint)((val-fx->parameters[0].min)*KEYSCALE/(fx->parameters[0].max-fx->parameters[0].min)+.5);
    }
    else {
      gint val=get_int_param(instance->parameter[0]);
      if (fx->parameters[0].min==fx->parameters[0].max) return (gint)fx->parameters[0].max;
      return (gint)((val-fx->parameters[0].min)*KEYSCALE/(fx->parameters[0].max-fx->parameters[0].min)+.5);
    }
  default:
    return 0;
  }
}


void livido_set_blend_factor(int hotkey) {
  livido_instance_t *instance;
  const livido_class_t *fx;
  int idx=key_to_instance[hotkey][key_modes[hotkey]];

  if (hotkey<0||idx==-1) return;

  if ((instance=livido_instance[idx])==NULL) return;
  if ((fx=instance->class)==NULL) return;
  if (fx->parameters==NULL) return;

  switch (fx->parameters[0].type) {
  case LVD_PARAM_NUMBER:
    if (fx->parameters[0].decimals) {
      set_float_param(instance->parameter[0],(gfloat)(fx->parameters[0].min+((gfloat)mainw->blend_factor/KEYSCALE*(fx->parameters[0].max-fx->parameters[0].min))));
    }
    else {
      set_int_param(instance->parameter[0],(gint)(fx->parameters[0].min+((gfloat)mainw->blend_factor/KEYSCALE*(fx->parameters[0].max-fx->parameters[0].min))+.5));
    }
    break;
  }
}

void livido_deinit_effect(int hotkey) {
  livido_class_t *fx;
  livido_instance_t *instance;
  gboolean is_resize=FALSE;

  if (hotkey<0) {
    is_resize=TRUE;
    hotkey=-hotkey;
  }

  if (hotkey>=1<<(num_livido_plugins-1)) return;

  fx=livido_fx[key_to_fx[hotkey][key_modes[hotkey]]];
  instance=livido_instance[key_to_instance[hotkey][key_modes[hotkey]]];
  if (my_in_channels[key_to_fx[hotkey][key_modes[hotkey]]][0].template==NULL) {
    // is generator
    if (mainw->playing_file>-1&&mainw->whentostop==STOP_ON_VID_END) {
      mainw->cancelled=1;
    }
    else {
      livido_generator_end (instance);
    }
    return;
  }
  else (*fx->deinit)(instance);

  if (mainw->rte_keys==hotkey&&mainw->rte_keys!=bg_generator_key) mainw->rte_keys=-1;
  if (mainw->last_grabable_effect==hotkey&&mainw->last_grabable_effect!=bg_generator_key) mainw->last_grabable_effect=-1;

  LIVIDO_FREE_INSTANCE(fx,instance);
  instance=NULL;
  livido_instance[key_to_instance[hotkey][key_modes[hotkey]]]=NULL;
  key_to_instance[hotkey][key_modes[hotkey]]=-1;


  if (my_in_channels[key_to_fx[hotkey][key_modes[hotkey]]][0].template!=NULL&&my_in_channels[key_to_fx[hotkey][key_modes[hotkey]]][1].template!=NULL) {
    mainw->num_tr_applied--;
    if (mainw->num_tr_applied<1&&!is_resize) {
      if (bg_gen_to_start!=-1) bg_gen_to_start=-1;
      else if (mainw->blend_file!=-1&&mainw->blend_file!=mainw->current_file&&mainw->files[mainw->blend_file]!=NULL&&mainw->files[mainw->blend_file]->clip_type==CLIP_TYPE_GENERATOR) {
	// all transitions off, so end the bg generator
	livido_generator_end (mainw->files[mainw->blend_file]->ext_src);
	mainw->blend_file=-1;
      }
    }
  }
}


void **store_livido_param_values (gint key) {
  // TODO - eventually use store_rfx_param_values, etc.
  // requires stabilisation of livido parameters
  livido_instance_t *inst;
  int i;
  void **store;
  const livido_class_t *fx;

  if (key_to_instance[key][key_modes[key]]==-1||(inst=livido_instance[key_to_instance[key][key_modes[key]]])==NULL) return NULL;
  fx=inst->class;
  if (fx->parameters==NULL) return NULL;
  for (i=0;fx->parameters[i].type!=LVD_PARAM_END;i++);
  store=g_malloc ((i+1)*(sizeof(void *)));

  for (i=0;fx->parameters[i].type!=LVD_PARAM_END;i++) {
    switch (fx->parameters[i].type) {
    case LVD_PARAM_NUMBER:
      if (fx->parameters[i].decimals) {
	store[i]=g_malloc (sizeof(float));
	memcpy (store[i],inst->parameter[i],sizeof(float));
      }
      else {
	store[i]=g_malloc (sizeof(int));
	memcpy (store[i],inst->parameter[i],sizeof(int));
      }
      break;
    default:
      store[i]=NULL;
    }
  } 
  return store;
}


void set_livido_params_from_store (gint key, void **store) {
  livido_instance_t *inst;
  int i;
  const livido_class_t *fx;

  if (store==NULL||key_to_instance[key][key_modes[key]]||(inst=livido_instance[key_to_instance[key][key_modes[key]]])==NULL) return;
  fx=inst->class;
  if (fx->parameters==NULL) return;
  for (i=0;fx->parameters[i].type!=LVD_PARAM_END;i++) {
    switch (fx->parameters[i].type) {
    case LVD_PARAM_NUMBER:
      if (fx->parameters[i].decimals) {
	memcpy (inst->parameter[i],store[i],sizeof(float));
      }
      else {
	memcpy (inst->parameter[i],store[i],sizeof(int));
      }
    }
  } 
}



void livido_params_store_free (gint key, void **store) {
  livido_instance_t *inst;
  int i;
  const livido_class_t *fx;

  if (store==NULL||key_to_instance[key][key_modes[key]]==-1||(inst=livido_instance[key_to_instance[key][key_modes[key]]])==NULL) return;
  fx=inst->class;
  for (i=0;fx->parameters[i].type!=LVD_PARAM_END;i++) {
    if (store[i]!=NULL) g_free (store[i]);
  } 
}

void livido_init_resize(gint rowstride) {
  // this is called a) at the start of playback, and b) when the frame size of
  // the current file changes
  void **param_store=NULL;
  int i;
  gint blend_file=mainw->blend_file;
  guint last_grabable_effect=mainw->last_grabable_effect;
  guint rte_keys=mainw->rte_keys;
  for (i=0;i<FX_KEYS-1;i++) {
    if (mainw->rte&(1<<i)) {
      // re-init only active effects
      if (i<num_livido_plugins&&my_in_channels[key_to_fx[i][key_modes[i]]][0].template!=NULL) {
	// and don't re-init generators
	if (livido_instance[key_to_instance[i][key_modes[i]]]->class->properties&LVD_PROPERTY_STATIC_CHANNEL_SIZES) {
	  // and only those which really, really need it
	  int which=i;
	  if (my_in_channels[key_to_fx[i][key_modes[i]]][1].template!=NULL) which=-which;
	  param_store=store_livido_param_values (i);
	  livido_deinit_effect(which);
	  mainw->rte_keys=rte_keys;
	  // reset all channel sizes
	  if (livido_init_effect(i,rowstride)) {
	    if (param_store!=NULL) {
	      set_livido_params_from_store (i,param_store);
	    }
	    if (i==(mainw->rte_keys-1)) livido_set_blend_factor(i);
	  }
	  if (param_store!=NULL) {
	    livido_params_store_free (i,param_store);
	    g_free (param_store);
	  }
	}
	else {
	  // just reset the channel sizes
	  set_channel_sizes (livido_instance[key_to_instance[i][key_modes[i]]]->in_channels,livido_instance[key_to_instance[i][key_modes[i]]]->out_channels);
	}
      }
    }
  }
  mainw->blend_file=blend_file;

  mainw->blend_factor=livido_get_blend_factor(mainw->rte_keys-1);

  mainw->last_grabable_effect=last_grabable_effect;
  mainw->reinit_effects=FALSE;
}




void livido_apply_effects(int height, int rowstride, guchar **in_pixel_data, livido_palette_t in_palette) {
  // for now, the in_palette will always come in as BGR888, this can be changed back and forth, but the final output should be BGR888 again
  
  // there could be a small optimisation, if mainw->ext_playback is TRUE
  // then the final output could be left as mainw->ext_palette
  // sometimes this can be UYVY

  // palettes need converting

  int i;
  guchar *out_pixel_data=g_malloc(height*rowstride);

  const livido_class_t *fx;
  livido_instance_t *instance;

  livido_timecode_t tc=LIVIDO_TIMECODE_UNDEFINED;

  livido_frame_t *ifp[2];
  livido_frame_t *ofp[1];
  
  livido_frame_t output_frame[]={{tc,0.0,0.0,out_pixel_data,NULL,NULL}};

  gint in_channs=1;
  livido_frame_t *bf_data=NULL;

  ifp[0]=(livido_frame_t *)g_malloc(sizeof(livido_frame_t));
  ifp[0]->timecode=tc;
  ifp[0]->x=0.;
  ifp[0]->y=0.;
  ifp[0]->pixel_data=in_pixel_data[0];
  ifp[0]->audio_data=NULL;
  ifp[0]->extra=NULL;

  if (in_pixel_data[1]!=NULL) {
    // transition
    ifp[1]=(livido_frame_t *)g_malloc(sizeof(livido_frame_t));
    ifp[1]->timecode=tc;
    ifp[1]->x=0.;
    ifp[1]->y=0.;
    ifp[1]->pixel_data=in_pixel_data[1];
    ifp[1]->audio_data=NULL;
    ifp[1]->extra=NULL;
    in_channs=2;
  }
  // plugin should know how many channels, thus how many elements in arrays

  // use inplace if we can
  
  for (i=0;i<FX_KEYS-1&&(livido_key_valid (i));i++) {
    if (mainw->rte&(1<<i)&&my_in_channels[key_to_fx[i][key_modes[i]]][0].template!=NULL) {
      // ignore generators
      fx=livido_fx[key_to_fx[i][key_modes[i]]];
      instance=livido_instance[key_to_instance[i][key_modes[i]]];

      if (my_in_channels[key_to_fx[i][key_modes[i]]][1].template==NULL) {
	ofp[0]=NULL;
	// single channel plugin
	if (!(fx->properties&LVD_PROPERTY_INPLACE)) {
	  ofp[0]=output_frame;
	}
	if (in_channs==2) {
	  // save ref to bg frame
	  bf_data=ifp[1];
	}
	ifp[1]=NULL;
	(*fx->process)(instance,ifp,ofp);
	if (!(fx->properties&LVD_PROPERTY_INPLACE)) {
	  memcpy(in_pixel_data[0],out_pixel_data,height*rowstride);
	}
	if (in_channs==2) {
	  ifp[1]=bf_data;
	}
      }
      else {
	ofp[0]=NULL;
	// two channel in effect
	if (in_channs==2) {
	  if (!(fx->properties&LVD_PROPERTY_INPLACE)) {
	    ofp[0]=output_frame; // this is our output
	  }
	  (*fx->process)(instance,ifp,ofp);
	  if (!(fx->properties&LVD_PROPERTY_INPLACE)) {
	    memcpy(in_pixel_data[0],out_pixel_data,height*rowstride);
	  }
	}
      }
    }
  }
  g_free (out_pixel_data);
  
  // note: ->pixel_data is not freed here, should be ref'ed elsewhere
  g_free(ifp[0]);
  if (in_channs==2) {
    g_free(ifp[1]);
  }
}



void livido_deinit_all(void) {
  // deinit all except generators
  // this is called on ctrl-0 or on shutdown
  int i;
  for (i=0;i<FX_KEYS-1&&livido_key_valid (i);i++) {
    if ((mainw->rte&(1<<i))&&my_in_channels[key_to_fx[i][key_modes[i]]][0].template!=NULL) {
      mainw->rte^=(1<<i);
      livido_deinit_effect(i);
    }
  }
}

void livido_end(void) {
  int i;
  mainw->num_tr_applied=0; // don't bother with all that
  livido_deinit_all();
  for (i=0;i<num_livido_plugins-num_modes;i++) {
    dlclose(livido_handles[i]);
    //g_free (livido_fx[i]);, etc
  }
}


void livido_key_setmode (gint key, gint newmode) {
  gint oldmode;
  gint blend_file;
  guint last_grabable_effect=mainw->last_grabable_effect;
  gshort whentostop=mainw->whentostop;
  gboolean was_rte_key=FALSE;
  gboolean was_started=FALSE;

  if (mainw->cancelled) return;

  if (key==0) {
    if ((key=mainw->rte_keys)==-1) return;
  }
  else key--;

  oldmode=key_modes[key];

  if (newmode==-1) {
    if (oldmode==MAX_MODES_PER_KEY||key_to_fx[key][oldmode+1]==-1) {
      newmode=0;
    }
    else {
      newmode=key_modes[key]+1;
    }
  }

  if (key_to_fx[key][newmode]==-1) return;


  if (mainw->playing_file==-1) {
    key_modes[ABS (key)]=newmode;
    return;
  }

  if (key==mainw->rte_keys) was_rte_key=TRUE;
  if (key_to_instance[ABS (key)][key_modes[ABS (key)]]!=-1&&livido_instance[key_to_instance[ABS (key)][key_modes [ABS (key)]]]!=NULL) was_started=TRUE;

  mainw->whentostop=NEVER_STOP; // when gen is off dont stop pb

  if (my_in_channels[key_to_fx[key][oldmode]][0].template!=NULL&&my_in_channels[key_to_fx[key][oldmode]][1].template!=NULL&&my_in_channels[key_to_fx[key][newmode]][0].template!=NULL&&my_in_channels[key_to_fx[key][newmode]][1].template!=NULL) {
    // transition --> transition, allow any bg generators to survive
    key=-key;
  }

  if (oldmode!=newmode) {
    blend_file=mainw->blend_file;

    if (my_in_channels[key_to_fx[ABS (key)][oldmode]][0].template!=NULL) {
      // not a generator
      if (was_started) {
	livido_deinit_effect (key);
      }
    }


    if (was_rte_key) {
      // leave our rte_keys unchanged
      mainw->rte_keys=ABS (key);
    }

    mainw->last_grabable_effect=last_grabable_effect;
    key_modes[ABS (key)]=newmode;

    mainw->blend_file=blend_file;

    if (was_started) {
      if (!livido_init_effect(key,mainw->frame_image==NULL?0:gdk_pixbuf_get_rowstride(mainw->frame_image))) {
	if (key<0) key=-key;
	if (mainw->rte&(1<<key)) mainw->rte^=(1<<key);
	return;
      }
    }
    mainw->whentostop=whentostop;
    mainw->cancelled=0;
    mainw->blend_factor=livido_get_blend_factor(ABS (key)-1);
  }
}


void livido_swap_fg_bg (void) {
  gint key=fg_generator_key;
  if (key!=-1) {
    if (mainw->rte&(1<<key)) mainw->rte^=(1<<key);
    fg_generator_clip=-1;
  }
  fg_generator_key=bg_generator_key;
  if (fg_generator_key!=-1) {
    mainw->rte|=(1<<fg_generator_key);
    fg_generator_clip=mainw->current_file;
  }
  bg_generator_key=key;
}


int livido_key_getmode (gint key) {
  return key_modes[key];
}


gchar *livido_key_getname (gint key, gint mode) {
  gchar *name=g_strdup ("");
  livido_class_t *fx;

  key--;

  if (key<0||key>FX_KEYS-1||mode<0||mode>MAX_MODES_PER_KEY) return name;

  if (key_to_fx[key][mode]==-1) return name;
  if ((fx=livido_fx[key_to_fx[key][mode]])==NULL) return name;
  if (fx->name==NULL) {
    g_free (name);
    name=g_strdup ("Unknown");
    return name;
  }
  g_free (name);
  name=g_strdup_printf ("%s",fx->name);
  return name;
}


int livido_getmodespk (void) {
  return MAX_MODES_PER_KEY;
}


#endif
