/*  
    Linux Dynamic Objects header
    
    This header should be included by filters and by host applications
    using them, in order to be able to use and implement video filters.
 */

/**
   @file livido.h

   @brief Linux Video Dynamic Objects header file
*/

/**
   @mainpage LiViDO :: Linux Video Dynamic Objects


   @section intro Introduction

During the last couple of years we have been observing a blooming
development in the field of realtime video software for the Linux/GNU
platform. An increasing number of artists and other users are
exploring the possibilities of several unique software packages for
video editing, mixing and effect processing.  The question arises here
how we (the developers) balance the growing number of features against
complexity, (re)usability and manageability of larger scale software
design.

As users, we choose our applications because their features fit our
purposes , as developers we serve the needs of the users who want an
increasing number of features added to our applications while we
should focus on flexible core engines and program frameworks.

An elegant solution to this problem is to provide the Free Software
community with  a flexible architecture that allows a shared pool of video
plugins. This brings a number of benefits:

- we can give designers of plugins the opportunity to develop their
(video) processing algorithms without the distraction of resolving
interface problems;
- we can share a set of unique plugins among a great number of programs;
- we can give the user the opportunity to control at least some aspects of
the features they want.

LiViDO design is based upon extensive research that has already gone
into applications such as PD/PDP , FreeJ, EffecTV , Veejay and other video
software applications.

The LiViDO architects will provide a lightweight, flexible API based
upon these established technologies. Currently, they are in the
process of creating a plug-in architecture useful for video editing
-mixing, and modular effect processing.

The API will deliberately be kept short and simple; to achieve
compatibility with a wide range of promising Linux video software
packages it is required to find the greatest common divisor in their logical
behaviour.

The API attempts to give programmers the ability to write simple plugin
video processors in C/C++ and link them dynamically with a range of
software packages (called hosts). It should be possible for any host and
any plugin to communicate through this interface.

Plugins will reside in shared object files (*.so in Linux), suitable
for dynamic linking by dlopen() and family. The file will provide a
number of plugin types that can be used to instantiate actual plugins
so they can be connected to perform these tasks.

Links to user groups and software supporting the LiViDO specification:

- http://www.piksel.no
- http://freej.dyne.org
- http://zwizwa.fartit.com/pd/pdp/overview.html
- http://veejay.sourceforge.net
- http://lives.sourceforge.net
- http://cvs.cinelerra.org/



   @section License
   
   LiViDO source code is free software; you can redistribute it and/or
   modify it under the terms of the GNU Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   LiViDO source code is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty
   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   Please refer to the GNU Public License for more details.
   
   You should have received a copy of the GNU Public License along
   with this source code; if not, write to: Free Software Foundation,
   Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   LiViDO is developed by:

   Denis "jaromil" Rojo - http://rastasoft.org

   Niels Elburg - http://veejay.sf.net

   Gabriel "Salsaman" Finch - http://lives.sourceforge.net

   Tom Schouten - http://zwizwa.fartit.com

   Andraz Tori - http://cvs.cinelerra.org

*/

#ifndef __LIVIDO_H__
#define __LIVIDO_H__

/** inttypes.h gives bit precision on integer types used
    (u)int(8,16,32,64)_t = (un)signed integer of a certain bit
    depth */
#include <inttypes.h>

/** buffer allocation routines from libc */
#include <stdlib.h>

/////////////////////////////////////////////////////////////////
/// LIVIDO API VERSION 1
/////////////////////////////////////////////////////////////////

/*


  	Fixes for livido generators

  Revision 1.9  salsaman 2004/10/20
  Palettes update by atori


  Revision 1.8  salsaman 2004/09/20
  Minor compilation fixes 


  Revision 1.7  salsaman 2004/08/20
  Add param flags, clean up some #defines 


  Revision 1.6  salsaman 2004/07
  Add support for channels


  Revision 1.5  2004/04/06 08:39:41  jaromil


  removed alloca() function,
  define to create instances from class,
  public parameters allocation and set.
  this update discards compatibility with previous revision.

  Revision 1.4  2004/03/30 18:57:20  jaromil


  now API includes class / instance mechanism
  see documentation in doxygen format inside livido.h

  also new working examples: plugin and host
  only libc needed: simple but works.

  comments welcome!

  Revision 1.3  2004/07/07 12:04:22  salsaman
  Updated documentation, fixed create instance

  Revision 1.2  2004/06/30 17:36:09  jaromil


  commit from trans.hackmeeting in MPhacklab:
  multiple input and output video channels,
  timecodes on frames,
  more palettes and mask-type palettes,
  more parameter types,
  general cleanup and review with Salsaman.

  Revision 1.1.1.1  2004/05/02 14:28:32  jaromil

  fresh import of LiViDO header core and examples,
  second review on sunday 2 may 2004



  Process me with doxygen to generate html documentation,
  or read thru this file, to explore the LiViDO API.

*/

#define LIVIDO_ID "$Id: plugins-livido.h,v 1.15 2004/11/03 02:27:26 gfinch Exp $"
///< last modification



#define LIVIDO_HEADER_VERSION 	1

/** 
    @defgroup palette Video palettes
    
    Livido Video Palettes
    
    a filter can specify more than one supported palette

    IMPORTANT: If applications use their own numbers, ask on the list to be assigned
               Do not put your personal definitions inside livido.h!
    GOOD PRACTICE: Palette names should be unambiguous.
                   We will add them when we have real applications using it, 
                   not because of theoretical ideas.
    
    @{ */

  typedef int livido_palette_t;
#define  LVD_PALETTE_FIRST_COL 1 /* First colour palette */
#define LVD_PALETTE_RGB888              1
#define LVD_PALETTE_RGBA8888            2
#define LVD_PALETTE_RGB161616           3
#define LVD_PALETTE_RGBA16161616        4
#define LVD_PALETTE_YUV888              5
#define LVD_PALETTE_YUVA8888            6
#define LVD_PALETTE_YUV161616           7
#define LVD_PALETTE_YUVA16161616        8
#define LVD_PALETTE_YUV422P             9 ///< 8 bits per channel!
#define LVD_PALETTE_YUV420P             10 ///< 8 bits per channel!
#define LVD_PALETTE_YUV444P             11 ///< 8 bits per channel!
#define LVD_PALETTE_RGBFLOAT            12
#define LVD_PALETTE_RGBAFLOAT           13
#define LVD_PALETTE_BGR888              14
#define LVD_PALETTE_YUYV8888            15
#define LVD_PALETTE_UYVY8888            16
#define LVD_PALETTE_RGB565              17
#define LVD_PALETTE_RGBX8888            18
#define  LVD_PALETTE_LAST_COL 18 /* Last colour palette */

/** Palettes >= 1024 are separated alpha channels */

#define  LVD_PALETTE_FIRST_ALPHA 1024
#define LVD_PALETTE_ALPHA1 1024
#define LVD_PALETTE_ALPHA2 1025
#define LVD_PALETTE_ALPHA4 1026
#define LVD_PALETTE_ALPHA8 1027
#define LVD_PALETTE_ALPHA16 1028
#define LVD_PALETTE_ALPHAFLOAT 1029
#define  LVD_PALETTE_LAST_ALPHA 1029

/** Palettes >= 2048 are custom palettes */

#define  LVD_PALETTE_FIRST_CUSTOM 2048

#define LVD_PALETTE_END         (0x0)    ///< marker for end of palette list

#define LVD_N_ALPHA_PALETTES (LVD_PALETTE_LAST_ALPHA-LVD_PALETTE_FIRST_ALPHA-1)
#define LVD_N_COL_PALETTES (LVD_PALETTE_LAST_COL-LVD_PALETTE_FIRST_COL-1)
#define LVD_N_PALETTES (LVD_N_ALPHA_PALETTES+LVD_N_COL_PALETTES) // excludes custom palettes

#define LIVIDO_PALETTE_IS_ALPHA(p) ( \
        (p&1024) ? 1 : 0 )

#define LIVIDO_PALETTE_IS_CUSTOM(p) ( \
        (p&2048) ? 1 : 0 )

///< check how many bytes is a palette (round up)
#define LIVIDO_PALETTE_BYTES(p) ( \
        (p == LVD_PALETTE_RGBAFLOAT)    ? 16 : \
        (p == LVD_PALETTE_RGBFLOAT)     ? 12 : \
        (p == LVD_PALETTE_RGBA16161616) || \
        (p == LVD_PALETTE_YUVA16161616) ? 8: \
        (p == LVD_PALETTE_RGB161616)    || \
        (p == LVD_PALETTE_YUV161616)    ? 6 : \
        (p == LVD_PALETTE_RGBA8888)     || \
        (p == LVD_PALETTE_RGBX8888)     || \
        (p == LVD_PALETTE_YUVA8888)     || \
        (p == LVD_PALETTE_ALPHAFLOAT)   || \
        (p == LVD_PALETTE_YUV444P)      ? 4 : \
        (p == LVD_PALETTE_RGB888)       || \
        (p == LVD_PALETTE_BGR888)       || \
        (p == LVD_PALETTE_YUV888)       ? 3 : \
        (p == LVD_PALETTE_YUV422P)      || \
        (p == LVD_PALETTE_YUYV8888)     || \
        (p == LVD_PALETTE_UYVY8888)     || \
        (p == LVD_PALETTE_ALPHA16)      || \
        (p == LVD_PALETTE_RGB565)       ? 2 : \
        (p == LVD_PALETTE_YUV420P)      ? 2 : \
        (p == LVD_PALETTE_ALPHA8)	? 1 : \
        (p == LVD_PALETTE_ALPHA1)?1 : \
        (p == LVD_PALETTE_ALPHA2)?1 : \
        (p == LVD_PALETTE_ALPHA4)?1 : \
        (p == LVD_PALETTE_ALPHA8)?1 : 0 )

///< check how many bits is a palette
#define LIVIDO_PALETTE_BITS(p) ( \
        (p == LVD_PALETTE_RGBAFLOAT)    ? 128 : \
        (p == LVD_PALETTE_RGBFLOAT)     ? 96 : \
        (p == LVD_PALETTE_RGBA16161616) || \
        (p == LVD_PALETTE_YUVA16161616) ? 64: \
        (p == LVD_PALETTE_RGB161616)    || \
        (p == LVD_PALETTE_YUV161616)    ? 48 : \
        (p == LVD_PALETTE_RGBA8888)     || \
        (p == LVD_PALETTE_RGBX8888)     || \
        (p == LVD_PALETTE_YUVA8888)     || \
        (p == LVD_PALETTE_ALPHAFLOAT)   || \
        (p == LVD_PALETTE_YUV444P)      ? 32 : \
        (p == LVD_PALETTE_RGB888)       || \
        (p == LVD_PALETTE_BGR888)       || \
        (p == LVD_PALETTE_YUV888)       ? 24 : \
        (p == LVD_PALETTE_YUV422P)      || \
        (p == LVD_PALETTE_YUYV8888)     || \
        (p == LVD_PALETTE_UYVY8888)     || \
        (p == LVD_PALETTE_ALPHA16)      ? 16 : \
        (p == LVD_PALETTE_RGB565)       ? 16 : \
        (p == LVD_PALETTE_YUV420P)      ? 12 : \
        (p == LVD_PALETTE_ALPHA8)	? 8: \
        (p == LVD_PALETTE_ALPHA1)?1: \
        (p == LVD_PALETTE_ALPHA2)?2: \
        (p == LVD_PALETTE_ALPHA4)?4: \
        (p == LVD_PALETTE_ALPHA8)?8:0 )


///< check if the palette is planar
#define LIVIDO_PALETTE_IS_PLANAR(p) ( \
        (p == LVD_PALETTE_YUV444P)      || \
        (p == LVD_PALETTE_YUV422P)      || \
	(p == LVD_PALETTE_YUV420P)	? 1 :0 )





/////////////////////////////////////////////////////////////////
/**
   @defgroup livido_channel_template_t Video channel template
   
   Channels templates are created by the plugin and offered to the host as part
   of the Plugin Class received when the livido_setup function is called on
   the plugin.

   These fields contain pointers to the first input and output
   channels respectively, which are the initial elements of arrays terminated by
   a channel with palettes==NULL.

   When the host receives the *in_channel and *out_channel channel template 
   arrays from the plugin, it must process them to see what channels are 
   available, and decide which it will use and how.

   The host should copy the channel templates to real channels, for each channel in turn, 
   the host should decide which of the palettes in the list it would like to use, remembering 
   that the list is in order of preference for the plugin. If none of the
   palettes are suitable in a channel, the host cannot use that
   channel. If channel that cannot be used is not optional, host cannot use this plugin.

   Assuming one of the palettes in the list is acceptible to the host,
   the host indicates this by setting the value of palette in livido_channel_t struct

   If a channel is optional, the plugin indicates this by setting the flag 
   LVD_CHNTMPL_HINT_OPTIONAL. The host can indicate not using the channel
   by setting disabled to 1 in channels that get passed to init

   The final task for the host is to set the channel width and
   height. If livido_channel_template_t.width is nonzero, the only possible 
   size that channel can take is denoted as width and heigth parameters of the template.
   
   If class property LVD_PROPERTY_STATIC_CHANNEL_SIZES is not set, host can 
   resize channels without reinitializing the instance.

   If class property LVD_PROPERTY_STATIC_CHANNEL_PALETTES is not set, host can 
   change palettes without reinitializing the instance.

   If the "same_as" field is set for any of the channels, the host
   should check that palette and size matches with the other channel, depending
   on the setting of LVD_CHNTMPL_SAME_AS_SIZE or 
   LVD_CHNTMPL_SAME_AS_PALETTE

*/

// CHANNEL TEMPLATE FLAGS

typedef uint32_t livido_chntmpl_flags_t;

// One or more of these flags can assist the host in setting up
// channel arrays

// hint to host that channel can be disabled	
#define LVD_CHNTMPL_HINT_OPTIONAL (1<<0)

// hint to host that channel is a mask
#define LVD_CHNTMPL_HINT_MASK (1<<1)

// tell host that same_as refers to size
#define LVD_CHNTMPL_SAME_AS_SIZE (1<<2)

// tell host that same_as refers to palettes
#define LVD_CHNTMPL_SAME_AS_PALETTE (1<<3)



/**

   @{ */

/**
   This structure is a template describing every single possible channel, it is
   held by the class and used to put create livido_channel_t structures passed
   back to plugin at initialization of the instance
   IMPORTANT: Host NEVER changes any part of this structure
*/

typedef struct {

  /** This field can be set to the name of the channel (e.g. "Input
      channel 1" if the plugin author likes. 
      IMPORTANT: Never set the name to point to NULL since that denotes the end of list
      GOOD PRACTICE: Avoid using words like "channel" inside the name, since it is implicit
  */
  char *name; ///< name of channel

  /** if channel is optional, host is allowed not to use it 
      IMPORTANT: Optional channels ALWAYS come at the end of the channel list
  */
  livido_chntmpl_flags_t flags; ///< flags/hints for this channel

  /** The plugin sets this to an array of palettes it can accept
      on this channel, in order of preference, terminated by a
      LVD_PALETTE_END (#defined as 0x0).
      The host writes selected value in livido_channel_t.palette
  */
  livido_palette_t *palettes;
  
  /** The plugin can use this field to tell the host that the settings
      of palette choice and size (width/height) must match that of 
      another channel. A normal
      use would be to mark the first output channel must match the
      first input channel. A positive value means "must match input
      channel n", a negative value means "must match output channel
      n". A setting of zero implies a free choice for palette and size
      settings. 
      
      IMPORTANT: both same_palette can only reference channels that 
      have serial numbers smaller then current (and in case of output 
      channels they can also reference all input channels, but input
      channels _canno_T reffer to output channels). 
      This way we avoid circular dependencies!
  */
  int same_as; ///< size and palette must be set to same as other channel


  /** Channel width and height. Leave them at 0 for host to chose
      them arbitrary. 
      
      IMPORTANT: You cannot just set height but leave width arbitrary 
      or vice versa. Both have to be zero or both have to be some other 
      value  
  */   
  int width; 
  int height;

  /** The plugin can use this field to tell the host that the settings
      of channel size must match that of another channel. A normal use
      would be to mark the first output channel must match the first
      input channel. A positive value means "must match input channel
      n", a negative value means "must match output channel n". A
      setting of zero implies a free choice for palette and size
      settings. The host should not change any of the "same_" values,
      otherwise the plugin may behave unpredictably. 
      
      IMPORTANT: both same_palette can only reference channels that 
      have serial numbers smaller then current (and in case of output 
      channels they can also reference all input channels, but input
      channels _canno_T refer to output channels). 
      This way we avoid circular dependencies!
      GOOD PRACTICE: same_palette and same_size should in most cases
      point to the same channel. It is very complicated for hosts
      to handle all this correctly, so be gentle.
  */

} livido_channel_template_t;


/// @}


// Values that denote end of channels list
#define LIVIDO_CHANNEL_TEMPLATE_END \
		{ \
		.name = NULL, \
		.flags = 0, \
		.palettes = NULL, \
		.same_as = 0, \
		.width = 0, \
		.height = 0, \
		}


// CHANNEL TEMPLATE FLAGS

// bitmap flags for channel, set by host

typedef uint32_t livido_channel_flags_t;

// channel is disabled (Set by host is channel is optional)	
#define LVD_CHANNEL_DISABLED (1<<0)



/**
   @defgroup livido_channel_t Video channel 
   
   This is description of single channel. It relates to the channel templates

   Channels are setup by host BEFORE plugin's init() function is called. Therefore
   plugin at the init stage knows what kind of data it will be recieving.

   GOOD PRACTICE: livido-library.c provides a rutine lvd_check_channel_layout()
		  that checks if channel layout passed to plugin corresponds to
		  class' template. Plugins should use it inside init() function.

   @{ */

typedef struct {
  /** the template that this channel is based upon 
  */ 
	livido_channel_template_t *template;

  /** If template of this channel allows it to be optional then setting
      disabled to 1 means it is not used 
  */	
	livido_channel_flags_t flags;             
 
  /** palette of the channel
      it is taken from one of the palettes provided by the template the channel was based on
  */
	livido_palette_t palette;   
	
  /** channel width and height in pixels
  */	
	int width; 
	int height; 

  /** shift_h and shift_v are  needed for 4:2:2, 4:2:0 modes and similar to know 
      how channels are aligned
  */
	
	int shift_h, shift_v; // this is needed for 4:2:2, 4:2:0 modes and similar to know how channels are aligned
 
  /** In packed modes rowstride[0] is the length of the row in bytes (including padding bytes)
      In planar modes every rowstride denotes the length of the plane in bytes!
     
     IMPORTANT: Every host has to set the rowstrides
     IMPORTANT: Every plugin has to honour the rowstrides
  */  	
	int rowstrides[4]; 

} livido_channel_t;

// Values that denote end of channels list
#define LIVIDO_CHANNEL_END \
		{ \
		.template = NULL, \
		.flags = 0, \
		.palette = LVD_PALETTE_END, \
		.width = 0, \
		.height = 0, \
                .shift_h = 0, \
                .shift_v = 0, \
                .rowstrides = {0,0,0,0} \
		}
	

/// @}




/////////////////////////////////////////////////////////////////






/**
   @defgroup Frame Format

   Frames in LiViDO are passed as structures. They hold the actual
   pixel data along with geometrical informations about it.
   
   They are used by the initialization function of filter Instances
   (see filter_public module) to describe the frame size of processed
   video. In case the filter has no LIVIDO_PROPERTY_RESIZABLE the
   sizes may not change during processing.
   
   Frames are also used for masks, which have a pixel data of 1 byte
   per pixel. In case masks are present for input or output of plugins,
   they are interleaved to the frames.

   
   @{ */
/////////////////////////////////////////////////////////////////

/**
   This structure stores the frame data along with informations on its
   geometry. It is used to pass frames to the process.
*/

typedef struct {
  uint64_t time_usec;         // microseconds since 'play start'
  int64_t frame_number; // frame number or 0 if undefined
} livido_timecode_t;


#define LIVIDO_TIMECODE_UNDEFINED \
{0, 0}


typedef struct {

  livido_timecode_t timecode;

  float x; ///< x position of the frame, 0 by default

  float y; ///< y position of the frame, 0 by default

  void *pixel_data; ///< pointer to the actual pixel data

  void *audio_data; ///< pointer to the audio data along the video (or NULL)
 
  void *extra; ///< extra pointer for future reference, unions (or NULL)

} livido_frame_t;

#define LIVIDO_FRAME_END \
{ LIVIDO_TIMECODE_END,0.,0.,NULL,NULL,NULL }
///< small define to mark the end of frames when the array is declared

/// @}






/////////////////////////////////////////////////////////////////
/**
   @defgroup instance Filter instance
   
    
   An instance is the buffer created by the filter at initialization
   
   @todo write something more here about the instance mechanism, that
   is what basically lets us use the same plugin multiple times...
   
   @{ */
/////////////////////////////////////////////////////////////////

/**
   This struct is allocated by the plugin init()
   
   it should be stored by the host and free'd at the end using
   the plugin function deinit()
   
   All variables inside the livido_instance are read and write, so they
   should be set by the host application before calling the alloca()
   function.

   The parameters array holds the real values for the parameters exposed
   by the filter, some of them require a new call to init() in case
   the livido_parameter_t::need_init flag is positive.  */


typedef struct {
  /** this is where the actual parameter values are stored, it must be
      allocated by the filter when the instance is created

      @todo macros to get/set parameters casted in proper types
  */
  struct _livido_class_t *class;

  void **parameter; ///< filter parameter values

  livido_channel_t *in_channels; ///< array of in channels

  livido_channel_t *out_channels; ///< array of out channels

  void *internal; ///< filter's private buffers

  char *error; ///< if there is an error is holding its description, NULL otherwise

} livido_instance_t;


/// @}




///////////////////////////////////////////////////////////////////
/** 
    @defgroup filter_public Filter class public functions
    
    The following public functions are found in the filter class,
    they are called by the host application to operate the filter.

    First the livido_setup symbol is called, which returns a
    structure which is the filter's class. The class will hold
    function pointers for the init() deinit() alloca() and process()
    functions.

    An host application should first setup the width, height and
    palette of the class, then it can proceed to create a new instance.
    
    The init() must be called in order to create an instance of
    the filter, which will be in turn passed to process() every
    time video input are processed into output.

    @{ */
///////////////////////////////////////////////////////////////////

/**
    Initialize the internal buffers and variables for a 
    instance of the filter, according to the parameters set.

    This function should be used on a freshly created filter instance,
    as well every time a parameter which needs init
    (see livido_parameter_t::need_init) is changed.

    It means that the init function must be REENTRANT.

    @brief initialize filter instance according to its parameters
    @param livido_instance_t filter instance to (re)initialize
    @param livido_frame_t descriptive frame instance to pass over geometrical informations
           this is important for filter that don't have LIVIDO_PROPERTY_RESIZABLE set.
    @return initialized filter instance
    */
typedef int (livido_init_f) (livido_instance_t*);

/**
   Free memory and cleanup internal variables.
   
   It must be called by host for de-initialization of instances when they
   are no more needed.
   
   @param filter instance
   @return error code
*/
typedef int (livido_deinit_f) (livido_instance_t*); 

/**
   Process instance->input frames to instance->output frames
   
   if LVD_PROPERTY_INPLACE is set , livido_data **out may be NULL
   
   @param env filter instance
   @param in input frame(s) to be processed, NULL terminated
   @param out output frame(s) processed, NULL terminated
          this is NULL if LIVIDO_PROPERTY_IS_INPLACE
   @return error code */
typedef int (livido_process_f) (livido_instance_t* instance,
				livido_frame_t **in, livido_frame_t **out);

///@}
















/** 
    @defgroup property Filter properties
    
    Properties of a filter define aspects of its functioning,
    they are flags set on the filter class, you can check them
    from the varialbe 'property' contained in it, i.e:
    if(property&LIVIDO_PROPERTY_*)

    Filters can have the following properties:

    @{ */


/**
   If this property is set, means that the filter can run in a (soft)
   realtime environment. The plugin must not do:
					 
   -# use malloc() free() or other memory managment functions that
   allocate/free data during processing.
   -# the plugin does not access (special) files in blocking
   operations
   -# the plugin does not do any colour space conversions but has
   internal routines for processing various colour spaces
   -# host is responsible for timing operations
*/
#define LVD_PROPERTY_REALTIME (1<<0x0) 
///< realtime filter property

/**
   If this property is set, the filter can do inplace operations then
   the plugin process() function will return NULL in *output and all
   the result will be in input (saves a memcpy on certain filters)
*/
#define LVD_PROPERTY_INPLACE (1<<0x1)
///< inplace filter property


#define LVD_PROPERTY_TRANSITION (1<<0x2)
///< transition filter property


// plugin knows how to scale the parameters when drawing previews/thumbnails 
// host can use scale_x and scale_y in this case 
#define LVD_PROPERTY_CAN_DO_SCALED (1<<0x3) 


// plugin knows how to operate on window of frame (adjusting parameters accordingly)
// window is denoted as window_x1, window_y1 for upper left corner of the window
// and full_x and full_y give the total size of the frame as it was before cropping
// lower right corner can be caluclated by adding window_x1 to channel width... and the same for height
#define LVD_PROPERTY_CAN_DO_WINDOWED (1<<0x4) 


// plugin can do automation by itself (keyframming)
#define LVD_PROPERTY_SELF_AUTOMATION (1<<0x5)


// plugin absolutely needs frames_per_second information, don't try to use it without
#define LVD_PROPERTY_FPS_NEEDED (1<<0x6) 


// should the init(..) be called by host after it resizes any channel (and before it calls process)
#define LVD_PROPERTY_STATIC_CHANNEL_SIZES (1<<0x7)

// should the init(..) be called by host after it changes palettes for any channel (and before it calls process)
#define LVD_PROPERTY_STATIC_CHANNEL_PALETTES (1<<0x8)

///@}







/////////////////////////////////////////////////////////////////
/**
   @defgroup parameter Filter parameters
   
   Livido Parameters are used by the plugin to describe the the
   controls of its filter, they should be parsed by the host loading
   the plugin to understand which values should be changed by the
   (user) interface.
    
   *PLEASE NOTE* That the real values are not stored here, but in the
   filter instance.
   
   @{ */
/////////////////////////////////////////////////////////////////



#define LVD_PARAM_BOOL 1	///< on/off switch
#define LVD_PARAM_NUMBER 2	///< double value
#define LVD_PARAM_RGB24	3///< color triplet for red green and blue (16 bits per channel)
#define LVD_PARAM_RGBA32 4	///< color triplet for red green blue and alpha (16 bits per channel)
#define LVD_PARAM_COORD2D 5	///< 2d cartesian coordinates
#define LVD_PARAM_STRING 6	///< string of maximum length = LIVIDO_STRING_SIZE
#define LVD_PARAM_STRINGLIST 7	///< strings of maximum length = LIVIDO_STRING_SIZE, one has to be picked
#define LVD_PARAM_POINTER 8	///< pointer to custom data buffer
#define LVD_PARAM_END 0	///< end of parameter declaration

#define LIVIDO_STRING_SIZE 	1024

typedef int32_t					lvd_type_bool;
typedef double					lvd_type_number;
typedef struct { uint16_t r,g,b; }	 	lvd_type_rgb24; // values for each channel are between 0 - 65535
typedef struct { uint16_t r,g,b,a; } 		lvd_type_rgba32;
typedef struct { double x,y; }			lvd_type_coord2d;
typedef char					lvd_type_string[LIVIDO_STRING_SIZE];
typedef lvd_type_string				(*lvd_type_stringlist);
typedef void *					lvd_type_pointer;


#define LIVIDO_PARAM_SIZE(p) ( \
	((p) == LVD_PARAM_BOOL)          ? sizeof(lvd_type_bool) : \
        ((p) == LVD_PARAM_NUMBER)     	 ? sizeof(lvd_type_number) : \
        ((p) == LVD_PARAM_RGB24)         ? sizeof(lvd_type_rgb24): \
        ((p) == LVD_PARAM_RGBA32)        ? sizeof(lvd_type_rgba32): \
        ((p) == LVD_PARAM_COORD2D)       ? sizeof(lvd_type_coord2d): \
        ((p) == LVD_PARAM_STRING)        ? sizeof(lvd_type_string): \
        ((p) == LVD_PARAM_STRINGLIST)    ? sizeof(lvd_type_stringlist): \
        ((p) == LVD_PARAM_POINTER)       ? sizeof(lvd_type_pointer) : 0)



#define LVD_PARAM_FLAG_NEEDS_INIT (1<<0x0) // need to re-init when value changes
#define LVD_PARAM_FLAG_TRANSITION (1<<0x1) // min is no transition, max is full

// This means if host uses keyframing, it should keyframe this parameter
#define LVD_PARAM_FLAG_KEYFRAMED (1<<0x4) 




/**
   It holds some variables that are interesting for the
   user interface, to render triggable realtime by users.

   @brief Struct to describe the public parameters of the plugin
*/
typedef struct {
const	char 	*name;	      ///< name
const	char	*description; ///< description
const   char    *hint;        ///< layout hints for the interface renderer

                              /// hints should use the keyword system, e.g:
                              /// layout|p1|fill|p2\n
                              /// special|aspect|1|2

const	int	type;	      ///< the parameter type (see livido_param_type_enum)
const   int     group;        ///< group parameters on positive values, zero means no group
const	float 	min;	      ///< minimum value
const	float	max;	      ///< maximum value
const   void	*def;         ///< default value
const   int     decimals;     ///< decimal places after the floating point
const   int	flags;
} livido_parameter_t;




#define LIVIDO_PARAM_END \
{ NULL, NULL, NULL, LVD_PARAM_END, 0, 0., 0., NULL, 0, 0 }
///< small define to mark the end of parameters when the array is declared

///@}




/**
   @defgroup class Filter class

   The class is the filter abstraction returned by the livido_setup()
   function, which is a symbol always found in every livido plugin.

   The setup function must be called only once, the first time the
   filter is loaded, and returns a structure holding pointers to
   all functions needed to operate a filter.

   Most notably the class will contain the name of the filter,
   its property flags, palettes supported and preferred and the
   init() deinit() and process() functions, while it will also
   hold the width and height geometry for the instance when is
   created by init().

   After calling the livido_setup(), an host application must call
   init() to obtain an instance of the class.
    
   @{ */

typedef struct _livido_class_t {

  char *name; ///< name of filter
  char *author; ///< author of filter
  char *description; ///< long text description
  int version; ///< filter version
  
  int livido_api_version; ///< version of the livido api used
  
  int properties;
  ///< filter properties
  
  livido_parameter_t		*parameters;
  ///< array of parameter descriptions, terminated with a LIVIDO_PARAM_END
  

  /* host should copy these channel arrays, set the palettes and optional flags,
     and then pass them into CREATE_INSTANCE, where they will be copied again for the plugin 
     host should free its own copy of channels */


  livido_channel_template_t *in_channel_templates;  // in_channels template
                                  ///< arrays terminated by a channel with 
                                  ///name==NULL

  livido_channel_template_t *out_channel_templates;  // out_channels template
                                   ///< arrays terminated by a channel with 
                                   ///name==NULL
 

  /* function pointers follow */
  livido_init_f 			*init;
  ///< init function creates a filter instance
  livido_deinit_f		*deinit; 
  ///< deinit function destroys a filter instance
  livido_process_f 		*process;
  ///< process function munches input videos in output
  
  void *extra; ///< extra data pointer reserved for future use
  
} livido_class_t;
#define LIVIDO_CLASS_END \
{ NULL,NULL,NULL,\
 0x0,0x0,0x0, \
   NULL,NULL,NULL,NULL,NULL,NULL,NULL }

/**
   the Setup is called by the host only the first time a plugin file is loaded,
   it doesn't needs to be called at every filter instance being created
   (that is the livido_init)

   It returns a null terminated array of filter classes made available
   by the plugin.

   @brief Setup the global environment of the plugin.
   @return NULL terminated array of filter classes available */
typedef livido_class_t *(livido_setup_f) (void);



/// @}


/**
   This define is a handy way to create new filter instances from a
   filter class, it might be used by the host application once a 
   plugin class is created.
   
   It executes the following operations:

   - allocate the instance buffer
   - allocate the buffers needed for the filter parameters
   - copy in_channels and out_channels supplied by host into the instance
   - set the filter parameters to the default values
   - set the internal filter instance buffer to NULL

   the above operations are necessary to create a breand new filter
   instance before setting it up and passing it to the init()

   @brief create a new clean filter instance from a filter class
   @param cl filter class
   @param inst pointer where to store the new filter instance */
#define LIVIDO_CREATE_INSTANCE(cl,inst) \
{ int c; size_t b; \
inst = malloc(sizeof(livido_instance_t)); \
if (cl->parameters) { \
 for(c=0;cl->parameters[c].type!=LVD_PARAM_END;c++); \
 inst->parameter = malloc(sizeof(void*) *c); \
 for(--c; c>=0; c--) { switch(cl->parameters[c].type) { \
	case LVD_PARAM_BOOL: b=sizeof(int); break; \
	case LVD_PARAM_NUMBER: \
 if (cl->parameters[c].decimals) b=sizeof(float); \
					      else b=sizeof(int); \
								    break; \
	  case LVD_PARAM_RGB24: b=sizeof(unsigned int)*3; break; \
	case LVD_PARAM_RGBA32: b=sizeof(unsigned int)*4; break; \
	case LVD_PARAM_COORD2D: b=sizeof(int)*2; break; \
	case LVD_PARAM_STRING: b=sizeof(char)*256; break; \
        default: b=sizeof(void*); \
	} \
  inst->parameter[c] = malloc(b); \
  memcpy(inst->parameter[c] , cl->parameters[c].def , b); \
  } \
      } \
inst->class=cl; \
inst->internal = NULL; }


#define LIVIDO_CREATE_INSTANCE_C(cl,inst,in_channs,out_channs) \
{ int c;\
livido_channel_t *in_array,*out_array; \
LIVIDO_CREATE_INSTANCE(cl,inst);\
for(c=0;in_channs[c].template!=NULL; c++);  \
in_array=(livido_channel_t *)malloc(++c*sizeof(livido_channel_t)); \
for(--c;c>=0;c--) \
  { \
    memcpy(&in_array[c],&in_channs[c],sizeof(livido_channel_t)); \
  } \
for(c=0;out_channs[c].template!=NULL; c++);  \
out_array=(livido_channel_t *)malloc(++c*sizeof(livido_channel_t)); \
for(--c;c>=0;c--) \
  { \
    memcpy(&out_array[c],&out_channs[c],sizeof(livido_channel_t)); \
  } \
inst->in_channels=in_array; \
inst->out_channels=out_array; \
}



#define LIVIDO_FREE_INSTANCE(cl,inst) \
{ int c; \
if (cl->parameters) { \
 for(c=0;cl->parameters[c].type!=LVD_PARAM_END;c++) \
	free( inst->parameter[c] ); \
 free (inst->parameter); \
 } \
free (inst->in_channels); \
free (inst->out_channels); \
free(inst); \
}



#endif // __LIVIDO_H__
